@@ -44,11 +44,15 @@ test_iconv_state (void)
 
   out = g_convert (in, -1, "UTF-8", "CP1255", 
 		   &bytes_read, &bytes_written, &error);
-  
+
+  if (error && error->code == G_CONVERT_ERROR_NO_CONVERSION)
+    return; /* silently skip if CP1255 is not supported, see bug 467707 */ 
+
   g_assert (error == NULL);
   g_assert (bytes_read == 5);
   g_assert (bytes_written == 10);
   g_assert (strcmp (out, expected) == 0);
+  g_free (out);
 }
 
 /* some tests involving "vulgar fraction one half" */
@@ -62,7 +66,7 @@ test_one_half (void)
   GError *error = NULL;  
 
   out = g_convert (in, -1, 
-		   "ISO8859-1", "UTF-8",
+		   "ISO-8859-1", "UTF-8",
 		   &bytes_read, &bytes_written,
 		   &error);
 
@@ -73,7 +77,7 @@ test_one_half (void)
   g_free (out);
 
   out = g_convert (in, -1, 
-		   "ISO8859-15", "UTF-8",
+		   "ISO-8859-15", "UTF-8",
 		   &bytes_read, &bytes_written,
 		   &error);
 
@@ -82,6 +86,7 @@ test_one_half (void)
   g_assert (bytes_written == 0);
   g_assert (out == NULL);
   g_clear_error (&error);
+  g_free (out);
 
   out = g_convert_with_fallback (in, -1, 
 				 "ISO8859-15", "UTF-8",
